-- A superclass for all player and AI-controlled units
import "CoreLibs/object"
import "CoreLibs/graphics"
import "CoreLibs/sprites"
import "CoreLibs/animation"
import "statemachine"

local gfx <const> = playdate.graphics
local geom <const> = playdate.geometry

class("Entity").extends(gfx.sprite)

function Entity:init(img, health, armor)
   Entity.super.init(self, img)
   self.type = "entity"
   self.health = health or 10
   self.armor = armor or 0
   self.introAnimator = nil

   -- For imagetable-based animation. see :animate()
   self.animation = nil

   -- movement direction, every update() the entity will move along this vector and return
   -- collision data to the subclass
   self.vector = geom.vector2D.new(0, 0)
   self:setCollideRect(0, 0, self:getSize())

   -- most entities will be enemies, so we configure this mask by default
   -- We don't set a collider mask because collision is a bit too variable
   -- (but we should always include 0x2 and handle player collisions)
   self:setGroupMask(0x4)

   -- state machine mapping
   -- this can be extended by subclasses if they need more states using
   -- StateMachine:addState()
   local states = {
      ["INIT"] = {
         main=self.runInit,
         incoming=self.onInit
      },
      ["INTRO"] = {
         main=self.runIntro,
         incoming=self.onIntro,
      },
      ["READY"] = {
         main=self.runReady,
         incoming=self.onReady,
      },
   }
   self.fsm = StateMachine.new({self}, states)
end

function Entity:damage(amount)
   if amount < self.armor then return end

   self.health = math.max(self.health - (amount - self.armor), 0)

   if self.health == 0 then
      self:remove()
   end
end

function Entity:add()
   Entity.super.add(self)
   if self.introAnimator then
      self.fsm:changeState("INTRO")
   else
      self.fsm:changeState("READY")
   end
end

function Entity:update()
   -- update state machine
   self.fsm:execute()

   if self.animation and self.animation:isValid() then
      -- TODO: consider efficiency of this vs implementing loop:draw() in a custom draw function
      self:setImage(self.animation:image())
   end
end

-- animate the entity based on the passed in imageTable.
-- duration is the length of *each* frame
-- If loop is true the animation will start over at the beginning
-- If reverse is true the animation will re-use the position in the *existing*
--   animation, but inverted. This assumes the new imgTable is the same number of
--   frames *and* is a valid reversal of the current animation
function Entity:animate(imgTable, duration, loop, reverse)
   local frame = 1

   -- reverse has no effect if there is not an active animation
   if reverse and self.animation and self.animation:isValid() then
      frame = #imgTable - self.animation.frame + 1
   end

   loop = loop or false
   self.animation = gfx.animation.loop.new(duration, imgTable, loop)
   self.animation.frame = frame
end

-- State machine-controlled functions

function Entity:onInit()
  -- stub for subclassing
end

function Entity:runInit()
  -- stub for subclassing
end

function Entity:onIntro()
   -- stub for subclassing
end

function Entity:runIntro()
   if self.introAnimator and not self.introAnimator:ended() then
      self:moveTo(self.introAnimator:currentValue())
   else
      self.fsm:changeState("READY")
   end
end

function Entity:onReady()
   -- stub for subclassing
end

function Entity:runReady()
   return select(3, self:moveWithCollisions(self.x + self.vector.dx, self.y + self.vector.dy))
end
